require 'puppet/resource_api'

Puppet::ResourceApi.register_type(
  name: 'dsc_sqlserverconfiguration',
  dscmeta_resource_friendly_name: 'SqlServerConfiguration',
  dscmeta_resource_name: 'MSFT_SqlServerConfiguration',
  dscmeta_module_name: 'SqlServerDsc',
  dscmeta_module_version: '12.0.0.0',
  docs: 'The DSC SqlServerConfiguration resource type.
         Automatically generated from version 12.0.0.0',
  features: ['simple_get_filter', 'canonicalize'],
  attributes: {
    name: {
      type:      'String',
      desc:      'Description of the purpose for this resource declaration.',
      behaviour: :namevar,
    },
    dsc_psdscrunascredential: {
      type: 'Optional[Struct[{ user => String[1], password => Sensitive[String[1]] }]]',
      desc: ' ',
      behaviour: :parameter,
      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'PSCredential',
      mof_is_embedded: true,
    },
    dsc_servername: {
      type: 'String',
      desc: 'The hostname of the SQL Server to be configured.',
      behaviour: :namevar,
      mandatory_for_get: true,
      mandatory_for_set: true,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_restartservice: {
      type: 'Optional[Boolean]',
      desc: 'Determines whether the instance should be restarted after updating the configuration option.',

      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'Boolean',
      mof_is_embedded: false,
    },
    dsc_instancename: {
      type: 'String',
      desc: 'Name of the SQL instance to be configured.',
      behaviour: :namevar,
      mandatory_for_get: true,
      mandatory_for_set: true,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_optionname: {
      type: 'String',
      desc: 'The name of the SQL configuration option to be checked.',
      behaviour: :namevar,
      mandatory_for_get: true,
      mandatory_for_set: true,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_optionvalue: {
      type: 'Integer[-2147483648, 2147483647]',
      desc: 'The desired value of the SQL configuration option.',

      mandatory_for_get: true,
      mandatory_for_set: true,
      mof_type: 'SInt32',
      mof_is_embedded: false,
    },
    dsc_restarttimeout: {
      type: 'Optional[Integer[0, 4294967295]]',
      desc: 'The length of time, in seconds, to wait for the service to restart. Default is 120 seconds.',

      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'UInt32',
      mof_is_embedded: false,
    },
  },
)
