require 'puppet/resource_api'

Puppet::ResourceApi.register_type(
  name: 'dsc_sqlserverendpoint',
  dscmeta_resource_friendly_name: 'SqlServerEndpoint',
  dscmeta_resource_name: 'MSFT_SqlServerEndpoint',
  dscmeta_module_name: 'SqlServerDsc',
  dscmeta_module_version: '12.0.0.0',
  docs: 'The DSC SqlServerEndpoint resource type.
         Automatically generated from version 12.0.0.0',
  features: ['simple_get_filter', 'canonicalize'],
  attributes: {
    name: {
      type:      'String',
      desc:      'Description of the purpose for this resource declaration.',
      behaviour: :namevar,
    },
    dsc_psdscrunascredential: {
      type: 'Optional[Struct[{ user => String[1], password => Sensitive[String[1]] }]]',
      desc: ' ',
      behaviour: :parameter,
      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'PSCredential',
      mof_is_embedded: true,
    },
    dsc_servername: {
      type: 'Optional[String]',
      desc: 'The host name of the SQL Server to be configured. Default value is $env:COMPUTERNAME.',

      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_instancename: {
      type: 'String',
      desc: 'The name of the SQL instance to be configured.',
      behaviour: :namevar,
      mandatory_for_get: true,
      mandatory_for_set: true,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_ensure: {
      type: "Optional[Enum['Present', 'Absent']]",
      desc: "If the endpoint should be present or absent. Default values is 'Present'.",

      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_endpointname: {
      type: 'String',
      desc: 'The name of the endpoint.',
      behaviour: :namevar,
      mandatory_for_get: true,
      mandatory_for_set: true,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_port: {
      type: 'Optional[Integer[0, 65535]]',
      desc: 'The network port the endpoint is listening on. Default value is 5022.',

      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'UInt16',
      mof_is_embedded: false,
    },
    dsc_ipaddress: {
      type: 'Optional[String]',
      desc: 'The network IP address the endpoint is listening on. Default the endpoint will listen on any valid IP address.',

      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'String',
      mof_is_embedded: false,
    },
  },
)
