require 'puppet/resource_api'

Puppet::ResourceApi.register_type(
  name: 'dsc_sqlsetup',
  dscmeta_resource_friendly_name: 'SqlSetup',
  dscmeta_resource_name: 'MSFT_SqlSetup',
  dscmeta_module_name: 'SqlServerDsc',
  dscmeta_module_version: '12.0.0.0',
  docs: 'The DSC SqlSetup resource type.
         Automatically generated from version 12.0.0.0',
  features: ['simple_get_filter', 'canonicalize'],
  attributes: {
    name: {
      type:      'String',
      desc:      'Description of the purpose for this resource declaration.',
      behaviour: :namevar,
    },
    dsc_rssvcaccountusername: {
      type: 'Optional[String]',
      desc: 'Output username for the Reporting Services service.',
      behaviour: :read_only,
      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_features: {
      type: 'Optional[String]',
      desc: 'SQL features to be installed.',

      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_asconfigdir: {
      type: 'Optional[String]',
      desc: 'Path for Analysis Services config.',

      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_sqluserdbdir: {
      type: 'Optional[String]',
      desc: 'Path for SQL database files.',

      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_sqltempdblogdir: {
      type: 'Optional[String]',
      desc: 'Path for SQL TempDB log files.',

      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_agtsvcaccount: {
      type: 'Optional[Struct[{ user => String[1], password => Sensitive[String[1]] }]]',
      desc: 'Service account for the SQL Agent service.',
      behaviour: :parameter,
      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'PSCredential',
      mof_is_embedded: true,
    },
    dsc_sqlbackupdir: {
      type: 'Optional[String]',
      desc: 'Path for SQL backup files.',

      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_suppressreboot: {
      type: 'Optional[Boolean]',
      desc: 'Suppresses reboot.',

      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'Boolean',
      mof_is_embedded: false,
    },
    dsc_productkey: {
      type: 'Optional[String]',
      desc: 'Product key for licensed installations.',

      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_ascollation: {
      type: 'Optional[String]',
      desc: 'Collation for Analysis Services.',

      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_setupprocesstimeout: {
      type: 'Optional[Integer[0, 4294967295]]',
      desc: 'The timeout, in seconds, to wait for the setup process to finish. Default value is 7200 seconds (2 hours). If the setup process does not finish before this time, and error will be thrown.',

      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'UInt32',
      mof_is_embedded: false,
    },
    dsc_securitymode: {
      type: "Optional[Enum['SQL', 'Windows']]",
      desc: "Security mode to apply to the SQL Server instance. 'SQL' indicates mixed-mode authentication while 'Windows' indicates Windows authentication. Default is Windows.",

      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_action: {
      type: "Optional[Enum['Install', 'InstallFailoverCluster', 'AddNode', 'PrepareFailoverCluster', 'CompleteFailoverCluster']]",
      desc: "The action to be performed. Default value is 'Install'.",

      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_agtsvcaccountusername: {
      type: 'Optional[String]',
      desc: 'Output username for the SQL Agent service.',
      behaviour: :read_only,
      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_failoverclustergroupname: {
      type: 'Optional[String]',
      desc: "The name of the resource group to create for the clustered SQL Server instance. Default is 'SQL Server (InstanceName)'.",

      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_assysadminaccounts: {
      type: 'Optional[Array[String]]',
      desc: 'Array of accounts to be made Analysis Services admins.',

      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'String[]',
      mof_is_embedded: false,
    },
    dsc_asservermode: {
      type: "Optional[Enum['MULTIDIMENSIONAL', 'TABULAR', 'POWERPIVOT']]",
      desc: 'The server mode for SQL Server Analysis Services instance. The default is to install in Multidimensional mode. Valid values in a cluster scenario are MULTIDIMENSIONAL or TABULAR. Parameter ASServerMode is case-sensitive. All values must be expressed in upper case.',

      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_agtsvcstartuptype: {
      type: "Optional[Enum['Automatic', 'Disabled', 'Manual']]",
      desc: 'Specifies the startup mode for SQL Server Sql Agent service.',

      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_installshareddir: {
      type: 'Optional[String]',
      desc: 'Installation path for shared SQL files.',

      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_sourcecredential: {
      type: 'Optional[Struct[{ user => String[1], password => Sensitive[String[1]] }]]',
      desc: "Credentials used to access the path set in the parameter 'SourcePath'.",
      behaviour: :parameter,
      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'PSCredential',
      mof_is_embedded: true,
    },
    dsc_rssvcstartuptype: {
      type: "Optional[Enum['Automatic', 'Disabled', 'Manual']]",
      desc: 'Specifies the startup mode for SQL Server Report service.',

      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_rssvcaccount: {
      type: 'Optional[Struct[{ user => String[1], password => Sensitive[String[1]] }]]',
      desc: 'Service account for Reporting Services service.',
      behaviour: :parameter,
      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'PSCredential',
      mof_is_embedded: true,
    },
    dsc_sqmreporting: {
      type: 'Optional[String]',
      desc: 'Enable customer experience reporting.',

      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_psdscrunascredential: {
      type: 'Optional[Struct[{ user => String[1], password => Sensitive[String[1]] }]]',
      desc: ' ',
      behaviour: :parameter,
      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'PSCredential',
      mof_is_embedded: true,
    },
    dsc_errorreporting: {
      type: 'Optional[String]',
      desc: 'Enable error reporting.',

      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_aslogdir: {
      type: 'Optional[String]',
      desc: 'Path for Analysis Services log files.',

      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_sqlsvcaccountusername: {
      type: 'Optional[String]',
      desc: 'Output username for the SQL service.',
      behaviour: :read_only,
      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_sqlsvcstartuptype: {
      type: "Optional[Enum['Automatic', 'Disabled', 'Manual']]",
      desc: 'Specifies the startup mode for SQL Server Sql Engine service.',

      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_instanceid: {
      type: 'Optional[String]',
      desc: 'SQL instance ID, if different from InstanceName.',

      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_failoverclusteripaddress: {
      type: 'Optional[Array[String]]',
      desc: 'Array of IP Addresses to be assigned to the clustered SQL Server instance.',

      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'String[]',
      mof_is_embedded: false,
    },
    dsc_astempdir: {
      type: 'Optional[String]',
      desc: 'Path for Analysis Services temp files.',

      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_updatesource: {
      type: 'Optional[String]',
      desc: 'Path to the source of updates to be applied during installation.',

      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_failoverclusternetworkname: {
      type: 'Optional[String]',
      desc: 'Host name to be assigned to the clustered SQL Server instance.',

      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_assvcstartuptype: {
      type: "Optional[Enum['Automatic', 'Disabled', 'Manual']]",
      desc: 'Specifies the startup mode for SQL Server Analysis service.',

      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_sqlcollation: {
      type: 'Optional[String]',
      desc: 'Collation for SQL.',

      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_sapwd: {
      type: 'Optional[Struct[{ user => String[1], password => Sensitive[String[1]] }]]',
      desc: "SA password, if SecurityMode is set to 'SQL'.",
      behaviour: :parameter,
      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'PSCredential',
      mof_is_embedded: true,
    },
    dsc_sourcepath: {
      type: 'Optional[String]',
      desc: 'The path to the root of the source files for installation. I.e and UNC path to a shared resource. Environment variables can be used in the path.',

      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_updateenabled: {
      type: 'Optional[String]',
      desc: 'Enabled updates during installation.',

      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_issvcstartuptype: {
      type: "Optional[Enum['Automatic', 'Disabled', 'Manual']]",
      desc: 'Specifies the startup mode for SQL Server Integration service.',

      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_assvcaccountusername: {
      type: 'Optional[String]',
      desc: 'Output username for the Analysis Services service.',
      behaviour: :read_only,
      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_asdatadir: {
      type: 'Optional[String]',
      desc: 'Path for Analysis Services data files.',

      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_sqluserdblogdir: {
      type: 'Optional[String]',
      desc: 'Path for SQL log files.',

      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_instancename: {
      type: 'String',
      desc: 'Name of the SQL instance to be installed.',
      behaviour: :namevar,
      mandatory_for_get: true,
      mandatory_for_set: true,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_instancedir: {
      type: 'Optional[String]',
      desc: 'Installation path for SQL instance files.',

      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_installsqldatadir: {
      type: 'Optional[String]',
      desc: 'Root path for SQL database files.',

      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_ftsvcaccountusername: {
      type: 'Optional[String]',
      desc: 'Output username for the Full Text service.',
      behaviour: :read_only,
      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_installsharedwowdir: {
      type: 'Optional[String]',
      desc: 'Installation path for x86 shared SQL files.',

      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_forcereboot: {
      type: 'Optional[Boolean]',
      desc: 'Forces reboot.',

      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'Boolean',
      mof_is_embedded: false,
    },
    dsc_sqltempdbdir: {
      type: 'Optional[String]',
      desc: 'Path for SQL TempDB files.',

      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_issvcaccount: {
      type: 'Optional[Struct[{ user => String[1], password => Sensitive[String[1]] }]]',
      desc: 'Service account for Integration Services service.',
      behaviour: :parameter,
      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'PSCredential',
      mof_is_embedded: true,
    },
    dsc_assvcaccount: {
      type: 'Optional[Struct[{ user => String[1], password => Sensitive[String[1]] }]]',
      desc: 'Service account for Analysis Services service.',
      behaviour: :parameter,
      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'PSCredential',
      mof_is_embedded: true,
    },
    dsc_sqlsysadminaccounts: {
      type: 'Optional[Array[String]]',
      desc: 'Array of accounts to be made SQL administrators.',

      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'String[]',
      mof_is_embedded: false,
    },
    dsc_browsersvcstartuptype: {
      type: "Optional[Enum['Automatic', 'Disabled', 'Manual']]",
      desc: 'Specifies the startup mode for SQL Server Browser service.',

      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_ftsvcaccount: {
      type: 'Optional[Struct[{ user => String[1], password => Sensitive[String[1]] }]]',
      desc: 'Service account for the Full Text service.',
      behaviour: :parameter,
      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'PSCredential',
      mof_is_embedded: true,
    },
    dsc_issvcaccountusername: {
      type: 'Optional[String]',
      desc: 'Output username for the Integration Services service.',
      behaviour: :read_only,
      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_asbackupdir: {
      type: 'Optional[String]',
      desc: 'Path for Analysis Services backup files.',

      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'String',
      mof_is_embedded: false,
    },
    dsc_sqlsvcaccount: {
      type: 'Optional[Struct[{ user => String[1], password => Sensitive[String[1]] }]]',
      desc: 'Service account for the SQL service.',
      behaviour: :parameter,
      mandatory_for_get: false,
      mandatory_for_set: false,
      mof_type: 'PSCredential',
      mof_is_embedded: true,
    },
  },
)
